# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import mxnet as mx

from gluonts.core.component import validated
from gluonts.itertools import prod
from gluonts.mx import Tensor
from gluonts.mx.block.feature import FeatureEmbedder
from gluonts.mx.block.scaler import MeanScaler, NOPScaler
from gluonts.mx.distribution import DistributionOutput
from gluonts.mx.util import weighted_average

from .trans_decoder import TransformerDecoder
from .trans_encoder import TransformerEncoder

LARGE_NEGATIVE_VALUE = -99999999


class TransformerNetwork(mx.gluon.HybridBlock):
    @validated()
    def __init__(
        self,
        encoder: TransformerEncoder,
        decoder: TransformerDecoder,
        history_length: int,
        context_length: int,
        prediction_length: int,
        distr_output: DistributionOutput,
        cardinality: List[int],
        embedding_dimension: int,
        lags_seq: List[int],
        scaling: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.history_length = history_length
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.scaling = scaling
        self.cardinality = cardinality
        self.embedding_dimension = embedding_dimension
        self.distr_output = distr_output

        assert len(set(lags_seq)) == len(
            lags_seq
        ), "no duplicated lags allowed!"
        lags_seq.sort()

        self.lags_seq = lags_seq

        self.target_shape = distr_output.event_shape

        with self.name_scope():
            self.proj_dist_args = distr_output.get_args_proj()
            self.encoder = encoder
            self.decoder = decoder
            self.embedder = FeatureEmbedder(
                cardinalities=cardinality,
                embedding_dims=[embedding_dimension for _ in cardinality],
            )

            if scaling:
                self.scaler = MeanScaler(keepdims=True)
            else:
                self.scaler = NOPScaler(keepdims=True)

    @staticmethod
    def get_lagged_subsequences(
        F,
        sequence: Tensor,
        sequence_length: int,
        indices: List[int],
        subsequences_length: int = 1,
    ) -> Tensor:
        """
        Returns lagged subsequences of a given sequence.

        Parameters
        ----------
        sequence : Tensor
            the sequence from which lagged subsequences should be extracted.
            Shape: (N, T, C).
        sequence_length : int
            length of sequence in the T (time) dimension (axis = 1).
        indices : List[int]
            list of lag indices to be used.
        subsequences_length : int
            length of the subsequences to be extracted.

        Returns
        --------
        lagged : Tensor
            a tensor of shape (N, S, C, I), where S = subsequences_length and
            I = len(indices), containing lagged subsequences. Specifically,
            lagged[i, j, :, k] = sequence[i, -indices[k]-S+j, :].
        """
        # we must have: sequence_length - lag_index - subsequences_length >= 0
        # for all lag_index, hence the following assert
        assert max(indices) + subsequences_length <= sequence_length, (
            "lags cannot go further than history length, found lag"
            f" {max(indices)} while history length is only {sequence_length}"
        )
        assert all(lag_index >= 0 for lag_index in indices)

        lagged_values = []
        for lag_index in indices:
            begin_index = -lag_index - subsequences_length
            end_index = -lag_index if lag_index > 0 else None
            lagged_values.append(
                F.slice_axis(
                    sequence, axis=1, begin=begin_index, end=end_index
                )
            )

        return F.stack(*lagged_values, axis=-1)

    def create_network_input(
        self,
        F,
        feat_static_cat: Tensor,  # (batch_size, num_features)
        past_time_feat: Tensor,  # (batch_size, num_features, history_length)
        past_target: Tensor,  # (batch_size, history_length, 1)
        past_observed_values: Tensor,  # (batch_size, history_length)
        future_time_feat: Optional[
            Tensor
        ],  # (batch_size, num_features, prediction_length)
        future_target: Optional[Tensor],  # (batch_size, prediction_length)
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Creates inputs for the transformer network.

        All tensor arguments should have NTC layout.
        """

        if future_time_feat is None or future_target is None:
            time_feat = past_time_feat.slice_axis(
                axis=1,
                begin=self.history_length - self.context_length,
                end=None,
            )
            sequence = past_target
            sequence_length = self.history_length
            subsequences_length = self.context_length
        else:
            time_feat = F.concat(
                past_time_feat.slice_axis(
                    axis=1,
                    begin=self.history_length - self.context_length,
                    end=None,
                ),
                future_time_feat,
                dim=1,
            )
            sequence = F.concat(past_target, future_target, dim=1)
            sequence_length = self.history_length + self.prediction_length
            subsequences_length = self.context_length + self.prediction_length

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags = self.get_lagged_subsequences(
            F=F,
            sequence=sequence,
            sequence_length=sequence_length,
            indices=self.lags_seq,
            subsequences_length=subsequences_length,
        )

        # scale is computed on the context length last units of the past target
        # scale shape is (batch_size, 1, *target_shape)
        _, scale = self.scaler(
            past_target.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
            past_observed_values.slice_axis(
                axis=1, begin=-self.context_length, end=None
            ),
        )
        embedded_cat = self.embedder(feat_static_cat)

        # in addition to embedding features, use the log scale as it can help
        # prediction too(batch_size, num_features + prod(target_shape))
        static_feat = F.concat(
            embedded_cat,
            (
                F.log(scale)
                if len(self.target_shape) == 0
                else F.log(scale.squeeze(axis=1))
            ),
            dim=1,
        )

        repeated_static_feat = static_feat.expand_dims(axis=1).repeat(
            axis=1, repeats=subsequences_length
        )

        # (batch_size, sub_seq_len, *target_shape, num_lags)
        lags_scaled = F.broadcast_div(lags, scale.expand_dims(axis=-1))

        # from (batch_size, sub_seq_len, *target_shape, num_lags)
        # to (batch_size, sub_seq_len, prod(target_shape) * num_lags)
        input_lags = F.reshape(
            data=lags_scaled,
            shape=(
                -1,
                subsequences_length,
                len(self.lags_seq) * prod(self.target_shape),
            ),
        )

        # (batch_size, sub_seq_len, input_dim)
        inputs = F.concat(input_lags, time_feat, repeated_static_feat, dim=-1)

        return inputs, scale, static_feat

    @staticmethod
    def upper_triangular_mask(F, d):
        mask = F.zeros_like(F.eye(d))
        for k in range(d - 1):
            mask = mask + F.eye(d, d, k + 1)
        return mask * LARGE_NEGATIVE_VALUE


class TransformerTrainingNetwork(TransformerNetwork):
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,
        past_time_feat: Tensor,
        past_target: Tensor,
        past_observed_values: Tensor,
        future_time_feat: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
    ) -> Tensor:
        """
        Computes the loss for training Transformer, all inputs tensors
        representing time series have NTC layout.

        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape,
            seq_len)
        future_time_feat : (batch_size, prediction_length, num_features)
        future_target : (batch_size, prediction_length, *target_shape)
        future_observed_values: (batch_size, prediction_length, *target_shape)

        Returns
        -------
        Loss with shape (batch_size, context + prediction_length, 1)
        """

        # create the inputs for the encoder
        inputs, scale, _ = self.create_network_input(
            F=F,
            feat_static_cat=feat_static_cat,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=future_time_feat,
            future_target=future_target,
        )

        enc_input = F.slice_axis(
            inputs, axis=1, begin=0, end=self.context_length
        )
        dec_input = F.slice_axis(
            inputs, axis=1, begin=self.context_length, end=None
        )

        # pass through encoder
        enc_out = self.encoder(enc_input)

        # input to decoder
        dec_output = self.decoder(
            dec_input,
            enc_out,
            self.upper_triangular_mask(F, self.prediction_length),
        )

        # compute loss
        distr_args = self.proj_dist_args(dec_output)
        distr = self.distr_output.distribution(distr_args, scale=scale)
        loss = distr.loss(future_target)

        # mask loss
        weighted_loss = weighted_average(
            F=F,
            x=loss,
            weights=future_observed_values,
            axis=1,
        )

        return weighted_loss.mean()


class TransformerPredictionNetwork(TransformerNetwork):
    @validated()
    def __init__(self, num_parallel_samples: int = 100, **kwargs) -> None:
        super().__init__(**kwargs)
        self.num_parallel_samples = num_parallel_samples

        # for decoding the lags are shifted by one, at the first time-step of
        # the decoder a lag of one corresponds to the last target value
        self.shifted_lags = [l - 1 for l in self.lags_seq]

    def sampling_decoder(
        self,
        F,
        static_feat: Tensor,
        past_target: Tensor,
        time_feat: Tensor,
        scale: Tensor,
        enc_out: Tensor,
    ) -> Tensor:
        """
        Computes sample paths by unrolling the LSTM starting with a initial
        input and state.

        Parameters
        ----------
        static_feat : Tensor
            static features. Shape: (batch_size, num_static_features).
        past_target : Tensor
            target history. Shape: (batch_size, history_length, 1).
        time_feat : Tensor
            time features. Shape:
            (batch_size, prediction_length, num_time_features).
        scale : Tensor
            tensor containing the scale of each element in the batch.
            Shape: (batch_size, ).
        enc_out: Tensor
            output of the encoder. Shape: (batch_size, num_cells)

        Returns
        --------
        sample_paths : Tensor
            a tensor containing sampled paths.
            Shape: (batch_size, num_sample_paths, prediction_length).
        """

        # blows-up the dimension of each tensor to batch_size *
        # self.num_parallel_samples for increasing parallelism
        repeated_past_target = past_target.repeat(
            repeats=self.num_parallel_samples, axis=0
        )
        repeated_time_feat = time_feat.repeat(
            repeats=self.num_parallel_samples, axis=0
        )
        repeated_static_feat = static_feat.repeat(
            repeats=self.num_parallel_samples, axis=0
        ).expand_dims(axis=1)
        repeated_enc_out = enc_out.repeat(
            repeats=self.num_parallel_samples, axis=0
        ).expand_dims(axis=1)
        repeated_scale = scale.repeat(
            repeats=self.num_parallel_samples, axis=0
        )

        future_samples = []

        # for each future time-units we draw new samples for this time-unit and
        # update the state
        for k in range(self.prediction_length):
            lags = self.get_lagged_subsequences(
                F=F,
                sequence=repeated_past_target,
                sequence_length=self.history_length + k,
                indices=self.shifted_lags,
                subsequences_length=1,
            )

            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_scaled = F.broadcast_div(
                lags, repeated_scale.expand_dims(axis=-1)
            )

            # from (batch_size * num_samples, 1, *target_shape, num_lags)
            # to (batch_size * num_samples, 1, prod(target_shape) * num_lags)
            input_lags = F.reshape(
                data=lags_scaled,
                shape=(-1, 1, prod(self.target_shape) * len(self.lags_seq)),
            )

            # (batch_size * num_samples, 1, prod(target_shape) * num_lags +
            # num_time_features + num_static_features)
            dec_input = F.concat(
                input_lags,
                repeated_time_feat.slice_axis(axis=1, begin=k, end=k + 1),
                repeated_static_feat,
                dim=-1,
            )

            dec_output = self.decoder(dec_input, repeated_enc_out, None, False)

            distr_args = self.proj_dist_args(dec_output)

            # compute likelihood of target given the predicted parameters
            distr = self.distr_output.distribution(
                distr_args, scale=repeated_scale
            )

            # (batch_size * num_samples, 1, *target_shape)
            new_samples = distr.sample()

            # (batch_size * num_samples, seq_len, *target_shape)
            repeated_past_target = F.concat(
                repeated_past_target, new_samples, dim=1
            )
            future_samples.append(new_samples)

        # reset cache of the decoder
        self.decoder.cache_reset()

        # (batch_size * num_samples, prediction_length, *target_shape)
        samples = F.concat(*future_samples, dim=1)

        # (batch_size, num_samples, *target_shape, prediction_length)
        return samples.reshape(
            shape=(
                (-1, self.num_parallel_samples)
                + self.target_shape
                + (self.prediction_length,)
            )
        )

    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,
        past_time_feat: Tensor,
        past_target: Tensor,
        past_observed_values: Tensor,
        future_time_feat: Tensor,
    ) -> Tensor:
        """
        Predicts samples, all tensors should have NTC layout.

        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape)
        future_time_feat : (batch_size, prediction_length, num_features)

        Returns predicted samples
        -------
        """

        # create the inputs for the encoder
        inputs, scale, static_feat = self.create_network_input(
            F=F,
            feat_static_cat=feat_static_cat,
            past_time_feat=past_time_feat,
            past_target=past_target,
            past_observed_values=past_observed_values,
            future_time_feat=None,
            future_target=None,
        )

        # pass through encoder
        enc_out = self.encoder(inputs)

        return self.sampling_decoder(
            F=F,
            past_target=past_target,
            time_feat=future_time_feat,
            static_feat=static_feat,
            scale=scale,
            enc_out=enc_out,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
from pathlib import Path

import torch
import torch.nn as nn

from gluonts.core.component import validated
from gluonts.dataset.field_names import FieldName
from gluonts.model.predictor import Predictor
from gluonts.torch.model.predictor import PyTorchPredictor
from gluonts.transform import InstanceSplitter, TestSplitSampler


class RandomNetwork(nn.Module):
    @validated()
    def __init__(
        self,
        prediction_length: int,
        context_length: int,
    ) -> None:
        super().__init__()
        assert prediction_length > 0
        assert context_length > 0
        self.prediction_length = prediction_length
        self.context_length = context_length
        self.net = nn.Linear(context_length, prediction_length)
        torch.nn.init.uniform_(self.net.weight, -1.0, 1.0)

    def forward(self, context):
        assert context.shape[-1] == self.context_length
        out = self.net(context)
        return out.unsqueeze(1)


def test_pytorch_predictor_serde():
    context_length = 20
    prediction_length = 5

    transformation = InstanceSplitter(
        target_field=FieldName.TARGET,
        is_pad_field=FieldName.IS_PAD,
        start_field=FieldName.START,
        forecast_start_field=FieldName.FORECAST_START,
        instance_sampler=TestSplitSampler(),
        past_length=context_length,
        future_length=prediction_length,
    )

    pred_net = RandomNetwork(
        prediction_length=prediction_length, context_length=context_length
    )

    predictor = PyTorchPredictor(
        prediction_length=prediction_length,
        input_names=["past_target"],
        prediction_net=pred_net,
        batch_size=16,
        input_transform=transformation,
        device="cpu",
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        predictor.serialize(Path(temp_dir))
        predictor_exp = Predictor.deserialize(Path(temp_dir))
    assert predictor == predictor_exp



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Tuple

import mxnet as mx

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .bijection import Bijection, InverseBijection
from .bijection_output import BijectionOutput
from .distribution import getF, softplus


class BoxCoxTransform(Bijection):
    r"""
    Implements Box-Cox transformation of a uni-variate random variable.

    The Box-Cox transformation of an observation :math:`z` is given by:

    .. math::

        BoxCox(z; \lambda_1, \lambda_2) = \begin{cases}
              ((z + \lambda_2)^{\lambda_1} - 1) / \lambda_1, \quad & \text{if }
              \lambda_1 \neq 0, \\
              \log (z + \lambda_2), \quad & \text{otherwise.}
          \end{cases}

    Here, :math:`\lambda_1` and :math:`\lambda_2` are learnable parameters.
    Note that the domain of the transformation is not restricted.

    For numerical stability, instead of checking :math:`\lambda_1` is exactly
    zero, we use the condition

    .. math::

        |\lambda_1| < tol\_lambda\_1

    for a pre-specified tolerance `tol_lambda_1`.

    Inverse of the Box-Cox Transform is given by

    .. math::

        BoxCox^{-1}(y; \lambda_1, \lambda_2) = \begin{cases}
            (y \lambda_1 + 1)^{(1/\lambda_1)} - \lambda_2, \quad & \text{if }
            \lambda_1 \neq 0, \\
            \exp (y) - \lambda_2, \quad & \text{otherwise.}
           \end{cases}

    **Notes on numerical stability:**

    1.  For the forward transformation, :math:`\lambda_2` must always be chosen
    such that

        .. math::

            z + \lambda_2 > 0.

        To achieve this one needs to know a priori the lower bound on the
        observations. This is set in `BoxCoxTransformOutput`,
        since :math:`\lambda_2` is learnable.

    2.  Similarly for the inverse transformation to work reliably, a sufficient
    condition is

        .. math::

            y \lambda_1 + 1 \geq 0,

        where :math:`y` is the input to the inverse transformation.

        This cannot always be guaranteed especially when :math:`y` is a sample
        from a transformed distribution. Hence we always
        truncate :math:`y \lambda_1 + 1` at zero.

        An example showing why this could happen in our case: consider
        transforming observations from the unit interval (0, 1) with
        parameters

        .. math::

            \begin{align}
                \lambda_1 = &\ 1.1, \\
                \lambda_2 = &\ 0.
            \end{align}

        Then the range of the transformation is (-0.9090, 0.0). If Gaussian is
        fit to the transformed observations and a sample is drawn from it,
        then it is likely that the sample is outside this range, e.g., when
        the mean is close to -0.9. The subsequent inverse transformation of
        the sample is not a real number anymore.

        >>> y = -0.91
        >>> lambda_1 = 1.1
        >>> lambda_2 = 0.0
        >>> (y * lambda_1 + 1) ** (1 / lambda_1) + lambda_2
        (-0.0017979146510711471+0.0005279153735965289j)

    Parameters
    ----------
    lambda_1
    lambda_2
    tol_lambda_1
        For numerical stability, treat `lambda_1` as zero if it is less than
        `tol_lambda_1`
    F
    """

    arg_names = ["box_cox.lambda_1", "box_cox.lambda_2"]

    @validated()
    def __init__(
        self,
        lambda_1: Tensor,
        lambda_2: Tensor,
        tol_lambda_1: float = 1e-2,
        F=None,
    ) -> None:
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.tol_lambda_1 = tol_lambda_1
        self.F = F if F else getF(lambda_1)

        # Addressing mxnet madness
        self._power = self.F.power if self.F == mx.nd else self.F.pow

    @property
    def args(self) -> List:
        r"""
        List: current values of the parameters
        """
        return [self.lambda_1, self.lambda_2]

    @property
    def event_dim(self) -> int:
        return 0

    @property
    def sign(self) -> Tensor:
        return 1.0

    def f(self, z: Tensor) -> Tensor:
        r"""
        Forward transformation of observations `z`

        Parameters
        ----------
        z
            observations

        Returns
        -------
        Tensor
            Transformed observations
        """
        F = self.F
        lambda_1 = self.lambda_1
        lambda_2 = self.lambda_2
        tol_lambda_1 = self.tol_lambda_1
        _power = self._power

        return F.where(
            condition=(F.abs(lambda_1).__ge__(tol_lambda_1).broadcast_like(z)),
            x=(_power(z + lambda_2, lambda_1) - 1.0) / lambda_1,
            y=F.log(z + lambda_2),
            name="Box_Cox_trans",
        )

    def f_inv(self, y: Tensor) -> Tensor:
        r"""
        Inverse of the Box-Cox Transform.

        Parameters
        ----------
        y
            Transformed observations

        Returns
        -------
        Tensor
            Observations
        """
        F = self.F
        lambda_1 = self.lambda_1
        lambda_2 = self.lambda_2
        tol_lambda_1 = self.tol_lambda_1
        _power = self._power

        # For numerical stability we truncate :math:`y * \lambda_1 + 1.0` at
        # zero.
        base = F.relu(y * lambda_1 + 1.0)

        return F.where(
            condition=(F.abs(lambda_1).__ge__(tol_lambda_1)).broadcast_like(y),
            x=_power(base, 1.0 / lambda_1) - lambda_2,
            y=F.exp(y) - lambda_2,
            name="Box_Cox_inverse_trans",
        )

    def log_abs_det_jac(self, z: Tensor, y: Tensor = None) -> Tensor:
        r"""
        Logarithm of the absolute value of the Jacobian determinant
        corresponding to the Box-Cox Transform is given by.

        .. math::
            \log \frac{d}{dz} BoxCox(z; \lambda_1, \lambda_2) = \begin{cases}
            \log (z + \lambda_2) (\lambda_1 - 1), \quad & \text{if } \lambda_1 \neq 0, \\
            -\log (z + \lambda_2), \quad & \text{otherwise.}
            \end{cases}

        Note that the derivative of the transformation is always non-negative.

        Parameters
        ----------
        z
            observations
        y
            not used

        Returns
        -------
        Tensor
        """  # noqa: E501
        F = self.F
        lambda_1 = self.lambda_1
        lambda_2 = self.lambda_2
        tol_lambda_1 = self.tol_lambda_1

        return F.where(
            condition=F.abs(lambda_1).__ge__(tol_lambda_1),
            x=F.log(z + lambda_2) * (lambda_1 - 1.0),
            y=-F.log(z + lambda_2),
            name="Box_Cox_trans_log_det_jac",
        )


class BoxCoxTransformOutput(BijectionOutput):
    bij_cls: type = BoxCoxTransform
    args_dim: Dict[str, int] = dict(zip(BoxCoxTransform.arg_names, [1, 1]))

    @validated()
    def __init__(self, lb_obs: float = 0.0, fix_lambda_2: bool = True) -> None:
        super().__init__()
        self.lb_obs = lb_obs
        self.fix_lambda_2 = fix_lambda_2

    def domain_map(self, F, *args: Tensor) -> Tuple[Tensor, ...]:
        lambda_1, lambda_2 = args
        if self.fix_lambda_2:
            lambda_2 = -self.lb_obs * F.ones_like(lambda_2)
        else:
            # This makes sure that :math:`z +  \lambda_2 > 0`, where :math:`z >
            # lb_obs`
            lambda_2 = softplus(F, lambda_2) - self.lb_obs * F.ones_like(
                lambda_2
            )

        # we squeeze the output since event_shape is ()
        return lambda_1.squeeze(axis=-1), lambda_2.squeeze(axis=-1)

    @property
    def event_shape(self) -> Tuple:
        return ()


class InverseBoxCoxTransform(InverseBijection):
    """
    Implements the inverse of Box-Cox transformation as a bijection.
    """

    arg_names = ["box_cox.lambda_1", "box_cox.lambda_2"]

    @validated()
    def __init__(
        self,
        lambda_1: Tensor,
        lambda_2: Tensor,
        tol_lambda_1: float = 1e-2,
        F=None,
    ) -> None:
        super().__init__(BoxCoxTransform(lambda_1, lambda_2, tol_lambda_1, F))

    @property
    def event_dim(self) -> int:
        return 0


class InverseBoxCoxTransformOutput(BoxCoxTransformOutput):
    bij_cls: type = InverseBoxCoxTransform

    args_dim: Dict[str, int] = dict(
        zip(InverseBoxCoxTransform.arg_names, [1, 1])
    )

    @property
    def event_shape(self) -> Tuple:
        return ()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from typing import List, NamedTuple
import numpy as np
import pandas as pd

from gluonts.dataset.common import ListDataset


class HierarchicalMetaData(NamedTuple):
    S: np.ndarray
    freq: str
    nodes: List


class HierarchicalTrainDatasets(NamedTuple):
    train: ListDataset
    test: ListDataset
    metadata: HierarchicalMetaData


def sine7(seq_length: int, prediction_length: int):
    x = np.arange(0, seq_length)

    # Bottom layer (4 series)
    amps = [0.8, 0.9, 1, 1.1]
    freqs = [1 / 20, 1 / 30, 1 / 50, 1 / 100]

    b = np.zeros((4, seq_length))
    for i, f in enumerate(freqs):
        omega = 0
        if i == 3:
            np.random.seed(0)
            omega = np.random.uniform(0, np.pi)  # random phase shift
        b[i, :] = amps[i] * np.sin(2 * np.pi * x * f + omega)

    # Aggregation matrix S
    S = np.array(
        [
            [1, 1, 1, 1],
            [1, 1, 0, 0],
            [0, 0, 1, 1],
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 0, 1],
        ],
    )

    Y = S @ b

    # Indices and timestamps
    index = pd.period_range(
        start=pd.Period("2020-01-01", freq="D"),
        periods=Y.shape[1],
        freq="D",
    )

    metadata = HierarchicalMetaData(
        S=S, freq=index.freqstr, nodes=[2, [2] * 2]
    )

    train_dataset = ListDataset(
        [
            {
                "start": index[0],
                "item_id": "all_items",
                "target": Y[:, :-prediction_length],
            }
        ],
        freq=index.freqstr,
        one_dim_target=False,
    )

    test_dataset = ListDataset(
        [{"start": index[0], "item_id": "all_items", "target": Y}],
        freq=index.freqstr,
        one_dim_target=False,
    )

    assert Y.shape[0] == S.shape[0]
    return HierarchicalTrainDatasets(
        train=train_dataset, test=test_dataset, metadata=metadata
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.
from typing import Optional

import pandas as pd

from .common import DataEntry
from .field_names import FieldName


def forecast_start(entry, time_axis: int = -1):
    return entry[FieldName.START] + entry[FieldName.TARGET].shape[time_axis]


def period_index(entry: DataEntry, freq=None) -> pd.PeriodIndex:
    if freq is None:
        freq = entry[FieldName.START].freq

    return pd.period_range(
        start=entry[FieldName.START],
        periods=entry[FieldName.TARGET].shape[-1],
        freq=freq,
    )


def to_pandas(entry: DataEntry, freq: Optional[str] = None) -> pd.Series:
    """
    Transform a dictionary into a pandas.Series object, using its "start" and
    "target" fields.

    Parameters
    ----------
    entry
        Dictionary containing the time series data.
    freq
        Frequency to use in the pandas.Series index.

    Returns
    -------
    pandas.Series
        Pandas time series object.
    """
    return pd.Series(
        entry[FieldName.TARGET],
        index=period_index(entry, freq=freq),
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import mxnet as mx
import numpy as np

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.mx import Tensor
from gluonts.mx.context import get_mxnet_context

from .representation import Representation


class HybridRepresentation(Representation):
    """
    A class representing a hybrid approach of combining multiple
    representations into a single representation. Representations will be
    combined by concatenating them on dim=-1.

    Parameters
    ----------
    representations
        A list of representations. Elements must be of type Representation.
    """

    @validated()
    def __init__(self, representations: List, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.representations = representations
        for representation in self.representations:
            self.register_child(representation)

    def initialize_from_dataset(
        self, input_dataset: Dataset, ctx: mx.Context = get_mxnet_context()
    ):
        for representation in self.representations:
            representation.initialize_from_dataset(input_dataset, ctx)

    def initialize_from_array(
        self, input_array: np.ndarray, ctx: mx.Context = get_mxnet_context()
    ):
        for representation in self.representations:
            representation.initialize_from_array(input_array, ctx)

    def hybrid_forward(
        self,
        F,
        data: Tensor,
        observed_indicator: Tensor,
        scale: Optional[Tensor],
        rep_params: List[Tensor],
        **kwargs,
    ) -> Tuple[Tensor, Tensor, List[Tensor]]:
        representation_list = []

        for representation in self.representations:
            representation_data, _, _ = representation(
                data,
                observed_indicator,
                scale,
                rep_params,
            )
            representation_list.append(representation_data)

        representation_agg = F.concat(*representation_list, dim=-1)

        if scale is None:
            scale = F.expand_dims(
                F.sum(data, axis=-1) / F.sum(observed_indicator, axis=-1), -1
            )

        return representation_agg, scale, []



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pandas as pd
import numpy as np
import torch
import pytest

from gluonts.dataset.pandas import PandasDataset
from gluonts.evaluation import make_evaluation_predictions
from gluonts.torch.model.estimator import PyTorchLightningEstimator
from gluonts.torch.model.simple_feedforward import SimpleFeedForwardEstimator
from gluonts.torch.model.deepar import DeepAREstimator
from gluonts.evaluation import Evaluator


FREQ = "D"
PRED_LENGTH = 5
TRAINER_KARGS = {"max_epochs": 1, "accelerator": "gpu", "devices": 1}


@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="only run the test on GPU"
)
@pytest.mark.parametrize(
    "model",
    [
        SimpleFeedForwardEstimator(
            prediction_length=PRED_LENGTH, trainer_kwargs=TRAINER_KARGS
        ),
        DeepAREstimator(
            freq=FREQ,
            prediction_length=PRED_LENGTH,
            trainer_kwargs=TRAINER_KARGS,
        ),
    ],
)
def test_torch_model_on_gpu(
    model: PyTorchLightningEstimator,
):
    n = 100
    date = "2015-04-07 00:00:00"

    df = pd.DataFrame(
        np.random.randn(n).astype(np.float32),
        index=pd.period_range(date, periods=n, freq=FREQ),
    )
    dataset = PandasDataset(df, target=0)

    predictor = model.train(dataset)

    forecast_it, ts_it = make_evaluation_predictions(
        dataset, predictor=predictor, num_samples=100
    )

    evaluator = Evaluator(quantiles=[0.1, 0.5, 0.9])
    agg_metrics, item_metrics = evaluator(
        ts_it,
        forecast_it,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
JSON Serialization/Deserialization
----------------------------------

The canonical way to do this is to define and `default` and `object_hook`
parameters to the json.dumps and json.loads methods. Unfortunately, due
to https://bugs.python.org/issue12657 this is not possible at the moment,
as support for custom NamedTuple serialization is broken.

To circumvent the issue, we pass the input value through custom encode
and decode functions that map nested object terms to JSON-serializable
data structures with explicit recursion.
"""

import json
from typing import Any, Optional

from ._base import decode, encode


def dump_json(o: Any, indent: Optional[int] = None) -> str:
    """
    Serializes an object to a JSON string.

    Parameters
    ----------
    o
        The object to serialize.
    indent
        An optional number of spaced to use as an indent.

    Returns
    -------
    str
        A string representing the object in JSON format.

    See Also
    --------
    load_json
        Inverse function.
    """
    return json.dumps(encode(o), indent=indent, sort_keys=True)


def load_json(s: str) -> Any:
    """
    Deserializes an object from a JSON string.

    Parameters
    ----------
    s
        A string representing the object in JSON format.

    Returns
    -------
    Any
        The deserialized object.

    See Also
    --------
    dump_json
        Inverse function.
    """
    return decode(json.loads(s))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.torch import DeepAREstimator
from gluonts.torch.distributions import StudentTOutput, NormalOutput
from gluonts.testutil.dummy_datasets import make_dummy_datasets_with_features


@pytest.mark.parametrize("datasets", [make_dummy_datasets_with_features()])
@pytest.mark.parametrize("distr_output", [StudentTOutput(), NormalOutput()])
def test_deepar_nonnegative_pred_samples(
    distr_output,
    datasets,
):
    estimator = DeepAREstimator(
        distr_output=distr_output,
        nonnegative_pred_samples=True,
        freq="D",
        prediction_length=3,
        trainer_kwargs={"max_epochs": 1},
    )

    dataset_train, dataset_test = datasets
    predictor = estimator.train(dataset_train)
    forecasts = list(predictor.predict(dataset_test))

    assert len(forecasts) == len(dataset_test)

    for forecast in forecasts:
        assert (forecast.samples >= 0).all()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pandas as pd
from packaging.version import Version

if Version(pd.__version__) <= Version("2.2.0"):
    S = "S"
    H = "H"
    M = "M"
    Q = "Q"
    Y = "A"
else:
    S = "s"
    H = "h"
    M = "ME"
    Q = "QE"
    Y = "YE"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest
from mxnet import nd

from gluonts.mx.block.encoder import HierarchicalCausalConv1DEncoder


@pytest.mark.parametrize("use_residual", [True, False])
@pytest.mark.parametrize("hybridize", [True, False])
def test_hierarchical_cnn_encoders(use_residual, hybridize) -> None:
    num_ts = 2
    ts_len = 10
    num_static_feat = 2
    num_dynamic_feat = 5

    test_data = nd.arange(num_ts * ts_len).reshape(shape=(num_ts, ts_len, 1))
    test_static_feat = nd.random.randn(num_ts, num_static_feat)
    test_dynamic_feat = nd.random.randn(num_ts, ts_len, num_dynamic_feat)

    chl_dim = [30, 30, 30]
    ks_seq = [3] * len(chl_dim)
    dial_seq = [1, 3, 9]

    cnn = HierarchicalCausalConv1DEncoder(
        dial_seq,
        ks_seq,
        chl_dim,
        use_residual,
        use_dynamic_feat=True,
        use_static_feat=True,
    )
    cnn.collect_params().initialize()

    if hybridize:
        cnn.hybridize()

    true_shape = (num_ts, ts_len, 31) if use_residual else (num_ts, ts_len, 30)

    assert (
        cnn(test_data, test_static_feat, test_dynamic_feat)[1].shape
        == true_shape
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Callable, List, Optional, Type

import mxnet.gluon.nn as nn
import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor


class FeatureEmbedder(nn.HybridBlock):
    """
    Embed a sequence of categorical features.

    Parameters
    ----------
    cardinalities
        cardinality for each categorical feature.

    embedding_dims
        number of dimensions to embed each categorical feature.

    dtype
        Data type of the embedded features.
    """

    @validated()
    def __init__(
        self,
        cardinalities: List[int],
        embedding_dims: List[int],
        dtype: Type = np.float32,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        assert (
            len(cardinalities) > 0
        ), "Length of `cardinalities` list must be greater than zero"
        assert len(cardinalities) == len(
            embedding_dims
        ), "Length of `cardinalities` and `embedding_dims` should match"
        assert all(
            [c > 0 for c in cardinalities]
        ), "Elements of `cardinalities` should be > 0"
        assert all(
            [d > 0 for d in embedding_dims]
        ), "Elements of `embedding_dims` should be > 0"

        self.__num_features = len(cardinalities)
        self.dtype = dtype

        def create_embedding(i: int, c: int, d: int) -> nn.Embedding:
            embedding = nn.Embedding(
                c, d, prefix=f"cat_{i}_embedding_", dtype=self.dtype
            )
            self.register_child(embedding)
            return embedding

        with self.name_scope():
            self.__embedders = [
                create_embedding(i, c, d)
                for i, (c, d) in enumerate(zip(cardinalities, embedding_dims))
            ]

    def hybrid_forward(self, F, features: Tensor) -> Tensor:
        """

        Parameters
        ----------
        F

        features
            Categorical features with shape: (N,T,C) or (N,C), where C is the
            number of categorical features.

        Returns
        -------
        concatenated_tensor: Tensor
            Concatenated tensor of embeddings with shape: (N,T,C) or (N,C),
            where C is the sum of the embedding dimensions for each categorical
            feature, i.e. C = sum(self.config.embedding_dims).
        """

        if self.__num_features > 1:
            # we slice the last dimension, giving an array of length
            # self.__num_features with shape (N,T) or (N)
            cat_feature_slices = F.split(
                features, axis=-1, num_outputs=self.__num_features
            )
        else:
            # F.split will iterate over the second-to-last axis if the last
            # axis is one
            cat_feature_slices = [features]

        return F.concat(
            *[
                embed(F.squeeze(cat_feature_slice, axis=-1))
                for embed, cat_feature_slice in zip(
                    self.__embedders, cat_feature_slices
                )
            ],
            dim=-1,
        )


class FeatureAssembler(nn.HybridBlock):
    """
    Assemble features into an MXNet tensor. Input features are distinguished
    based on the following criteria:

    - static (time-independent) features vs dynamic (that is, time-dependent)
    - categorical vs real-valued features.

    Dynamic features have shape `(N, T, C)` and static features have shape
    `(N, C)`, where

        - `N` is the number of elements in the processed batch,
        - `T` is the time dimension,
        - `C` is the number of features.

    If multiple feature types are used, the :class:`FeatureAssembler` will
    assume that the N and T dimensions are the same for all passed arguments.

    Categorical features can be optionally embedded using trained embedding
    layers via nested :class:`FeatureEmbedder` components.

    >>> embed_static = FeatureEmbedder(
    ...     cardinalities=[2],
    ...     embedding_dims=[3],
    ...     prefix='embed_static_',
    ... )
    >>> embed_dynamic = FeatureEmbedder(
    ...     cardinalities=[5, 5],
    ...     embedding_dims=[6, 9],
    ...     prefix='embed_dynamic_',
    ... )

    The above snippet with four :class:`nn.Embedding` corresponding to the one
    static and two dynamic categorical features. The `(input_dim, output_dim)`
    of these layers are going to be `(2, 3)`, `(5, 6)`, and `(5, 9)`. The
    created `assemble_feature` instance will not handle real-valued features.

    The subset of feature types to be used by the :class:`FeatureAssembler`
    instance is determined using corresponding constructor parameters. Here is
    an example that constructs a feature assembler consuming only real-valued
    features.

    >>> N, T = 50, 168
    >>> assemble_feature = FeatureAssembler(
    ...     T=T,
    ...     # use_static_cat=True,
    ...     # use_static_real=False,
    ...     # use_dynamic_cat=True,
    ...     # use_dynamic_real=False,
    ...     embed_static=embed_static,
    ...     embed_dynamic=embed_dynamic
    ... )

    When the `__call__`, `forward`, or `hybrid_forward` methods of
    a :class:`FeatureAssembler` are called, we always have to pass a full set
    of features. Missing features are represented as zero tensors with a
    suitable shape.

    For example,

    >>> import mxnet as mx
    >>> feat_static_cat = mx.nd.random.uniform(0, 2, shape=(N, 1)).floor()
    >>> feat_dynamic_cat = mx.nd.random.uniform(
        0, 5, shape=(N, 168, 2)
    ).floor()
    >>> feat_static_real = mx.nd.zeros(shape=(N, 1,)) # empty feature
    >>> feat_dynamic_real = mx.nd.zeros(shape=(N, T, 1,)) # empty feature

    After initializing the embedder parameters to one and instantiating some
    random `static_cat` and `dynamic_cat` vectors,

    >>> assemble_feature.collect_params().initialize(mx.initializer.One())

    one can do a forward pass as follows.

    >>> assembled_feature = assemble_feature(
        feat_static_cat, feat_static_real, feat_dynamic_cat, feat_dynamic_real
    )
    >>> assembled_feature.shape
    (50, 168, 20)
    >>>

    However, relative order of `static_cat` and `dynamic_cat` in the call above
    is determined by the fact that `use_static_cat` is defined before
    `use_dynamic_cat` in the class constructor.
    """

    @validated()
    def __init__(
        self,
        T: int,
        use_static_cat: bool = False,
        use_static_real: bool = False,
        use_dynamic_cat: bool = False,
        use_dynamic_real: bool = False,
        embed_static: Optional[FeatureEmbedder] = None,
        embed_dynamic: Optional[FeatureEmbedder] = None,
        dtype: Type = np.float32,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        assert T > 0, "The value of `T` should be > 0"

        self.T = T
        self.dtype = dtype
        self.use_static_cat = use_static_cat
        self.use_static_real = use_static_real
        self.use_dynamic_cat = use_dynamic_cat
        self.use_dynamic_real = use_dynamic_real
        self.embed_static: Callable[[Tensor], Tensor] = embed_static or (
            lambda x: x
        )
        self.embed_dynamic: Callable[[Tensor], Tensor] = embed_dynamic or (
            lambda x: x
        )

    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,
        feat_static_real: Tensor,
        feat_dynamic_cat: Tensor,
        feat_dynamic_real: Tensor,
    ) -> Tensor:
        processed_features = [
            self.process_static_cat(F, feat_static_cat),
            self.process_static_real(F, feat_static_real),
            self.process_dynamic_cat(F, feat_dynamic_cat),
            self.process_dynamic_real(F, feat_dynamic_real),
        ]

        return F.concat(*processed_features, dim=-1)

    def process_static_cat(self, F, feature: Tensor) -> Tensor:
        feature = self.embed_static(feature.astype(self.dtype))
        return F.tile(feature.expand_dims(axis=1), reps=(1, self.T, 1))

    def process_dynamic_cat(self, F, feature: Tensor) -> Tensor:
        return self.embed_dynamic(feature.astype(self.dtype))

    def process_static_real(self, F, feature: Tensor) -> Tensor:
        return F.tile(feature.expand_dims(axis=1), reps=(1, self.T, 1))

    def process_dynamic_real(self, F, feature: Tensor) -> Tensor:
        return feature



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path

import click

from gluonts.dataset.common import FileDataset
from gluonts.gluonts_tqdm import tqdm

from . import ArrowWriter, ParquetWriter


@click.group()
def cli():
    pass


@cli.command()
@click.argument("dataset", type=click.Path(exists=True))
@click.argument("out", type=click.Path())
@click.option("--freq", required=True)
@click.option(
    "--type", "type_", type=click.Choice(["arrow", "feather", "parquet"])
)
@click.option("--stream", default=False)
@click.option("--compression")
def write(dataset, out, freq, type_, stream, compression):
    out = Path(out)

    if type_ is None:
        if out.suffix in [".feather", ".arrow"]:
            type_ = "arrow"
        elif out.suffix == ".parquet":
            type_ = "parquet"
        else:
            raise click.UsageError(f"Unsupported suffix {out.suffix}.")

    if type_ == "arrow":
        writer = ArrowWriter(stream=stream, compression=compression)
    else:
        writer = ParquetWriter()

    writer.write_to_file(tqdm(FileDataset(dataset, freq)), out)


if __name__ == "__main__":
    cli()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import copy
import dataclasses

import numpy as np

from toolz import valmap


@dataclasses.dataclass
class AxisView:
    data: np.ndarray
    axis: int

    def __getitem__(self, index):
        slices = [slice(None)] * self.data.ndim
        slices[self.axis] = index

        return self.data[tuple(slices)]

    def __len__(self):
        return self.data.shape[self.axis]


def pad_axis(
    a: np.ndarray, *, axis: int = 0, left: int = 0, right: int = 0, value=0
) -> np.ndarray:
    """Similar to ``np.pad``, but pads only a single axis using `left` and
    `right` parameters.
    ::
        >>> pad_axis([1, 2, 3, 4], left=2, right=3)
        array([0, 0, 1, 2, 3, 4, 0, 0, 0])
    """
    a = np.array(a)

    pad_width = [(0, 0)] * a.ndim
    pad_width[axis] = (left, right)
    return np.pad(a, pad_width, constant_values=value)


def _replace(obj, **kwargs):
    """
    Copy and replace dataclass instance.

    Compared to ``dataclasses.replace`` this first creates a copy where each
    field in the object is copied. Thus, each field of the returned object is
    different from the source object.
    """

    clone = object.__new__(obj.__class__)
    clone.__dict__ = valmap(copy.copy, obj.__dict__)

    return dataclasses.replace(clone, **kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from operator import add
from pathlib import Path
from typing import NamedTuple

import numpy as np
import pandas as pd
import pytest

from gluonts.core.component import equals, equals_list
from gluonts.core import serde


class Span(NamedTuple):
    path: Path
    line: int


class BestEpochInfo(NamedTuple):
    params_path: Path
    epoch_no: int
    metric_value: float


# Example Instances
# -----------------

best_epoch_info = BestEpochInfo(
    params_path=Path("foo/bar"), epoch_no=1, metric_value=0.5
)


numpy_array = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64)

list_container = [
    best_epoch_info,
    42,
    0.7,
    "fx",
    numpy_array,
]

set_container = {best_epoch_info, 42, 0.7, "fx"}

dict_container = dict(best_epoch_info=best_epoch_info)

simple_types = [
    1,
    42.0,
    "Oh, Romeo",
    np.int32(33),
    np.float64(3.1415),
]  # float('nan')

complex_types = [
    Path("foo/bar"),
    best_epoch_info,
    numpy_array,
]

container_types = [list_container, dict_container, set_container]

examples = simple_types + complex_types + container_types  # type: ignore


def check_equality(expected, actual) -> bool:
    if isinstance(expected, set):
        # Sets are serialized as lists — we check if they have the same elements
        return equals_list(
            sorted(expected, key=hash), sorted(actual, key=hash)
        )
    elif np.issubdtype(type(expected), np.integer):
        # Integer types are expected to be equal exactly
        return np.equal(expected, actual)
    elif np.issubdtype(type(expected), np.inexact):
        # Floating point types are expected to be equal up a certain digit, as specified in np.isclose
        return np.allclose(expected, actual)
    else:
        return equals(expected, actual)


@pytest.mark.parametrize("e", examples)
def test_json_serialization(e) -> None:
    expected, actual = e, serde.load_json(serde.dump_json(e))
    assert check_equality(expected, actual)


def test_timestamp_encode_decode() -> None:
    now = pd.Timestamp.now()
    assert now == serde.decode(serde.encode(now))


def test_string_escape() -> None:
    assert serde.load_json(serde.dump_json(r"a\b")) == r"a\b"


def test_serde_fq():
    add_ = serde.decode(serde.encode(add))
    assert add_(1, 2) == 3

    def foo():
        pass

    with pytest.raises(Exception):
        serde.encode(foo)


def test_serde_partial():
    add_1 = partial(add, 1)

    add_1_ = serde.decode(serde.encode(add_1))

    assert add_1_(2) == 3


class X(serde.Stateless):
    def m(self):
        return 42


def test_serde_method():
    x = X()

    m = serde.decode(serde.encode(x.m))

    assert m() == 42


def test_np_str_dtype():
    a = np.array(["foo"])
    serde.decode(serde.encode(a.dtype)) == a.dtype


@pytest.mark.parametrize(
    "obj",
    [
        {"__kind__": 42, "class": cls_str}
        for cls_str in [
            "builtins.eval",
            "builtins.exec",
            "builtins.compile",
            "builtins.open",
            "builtins.input",
            "eval",
            "exec",
            "compile",
            "open",
            "input",
        ]
    ],
)
def test_decode_disallow(obj):
    with pytest.raises(ValueError):
        serde.decode(obj)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial

import pytest

from gluonts.mx.model.deepstate import DeepStateEstimator
from gluonts.mx.trainer import Trainer
from gluonts.testutil.dummy_datasets import make_dummy_datasets_with_features

common_estimator_hps = dict(
    freq="D",
    prediction_length=3,
    trainer=Trainer(epochs=1, num_batches_per_epoch=1, hybridize=True),
    past_length=10,
    add_trend=True,
)


@pytest.mark.parametrize(
    "estimator, datasets",
    [
        # No features
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[1],
                use_feat_static_cat=False,
            ),
            make_dummy_datasets_with_features(),
        ),
        # Single static categorical feature
        (
            partial(
                DeepStateEstimator, **common_estimator_hps, cardinality=[5]
            ),
            make_dummy_datasets_with_features(cardinality=[5]),
        ),
        # Multiple static categorical features
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[3, 10, 42],
            ),
            make_dummy_datasets_with_features(cardinality=[3, 10, 42]),
        ),
        # Multiple static categorical features for one of which cardinality = 1
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[3, 1, 42],
            ),
            make_dummy_datasets_with_features(cardinality=[3, 1, 42]),
        ),
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[1],
                use_feat_static_cat=False,
                use_feat_dynamic_real=True,
            ),
            make_dummy_datasets_with_features(num_feat_dynamic_real=1),
        ),
        # Multiple dynamic real feature
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[1],
                use_feat_static_cat=False,
                use_feat_dynamic_real=True,
            ),
            make_dummy_datasets_with_features(num_feat_dynamic_real=3),
        ),
        # Multiple dynamic real feature (ignored)
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[1],
                use_feat_static_cat=False,
            ),
            make_dummy_datasets_with_features(num_feat_dynamic_real=3),
        ),
        # Both static categorical and dynamic real features
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                use_feat_dynamic_real=True,
                use_feat_static_cat=True,
                cardinality=[3, 10, 42],
            ),
            make_dummy_datasets_with_features(
                cardinality=[3, 10, 42], num_feat_dynamic_real=3
            ),
        ),
    ],
)
def test_deepstate_smoke(estimator, datasets):
    # TODO: pass `dtype` below once you add `dtype` support to `DeepStateEstimator`
    estimator = estimator()
    dataset_train, dataset_test = datasets
    predictor = estimator.train(dataset_train)
    forecasts = list(predictor.predict(dataset_test))
    assert len(forecasts) == len(dataset_test)


# This test makes sure that `DeepStateEstimator` raises exception when it is misused.
#   * cardinality is not passed explicitly
#   * product of cardinalities is not greater than 1 even though `use_feat_static_cat == True`
@pytest.mark.parametrize(
    "estimator, datasets",
    [
        # Static categorical feature ignored
        (
            partial(DeepStateEstimator, **common_estimator_hps),
            make_dummy_datasets_with_features(cardinality=[3]),
        ),
        # Static categorical features are ignored by wrongly setting cardinality = [1, 1, 1]
        (
            partial(
                DeepStateEstimator,
                **common_estimator_hps,
                cardinality=[1, 1, 1],
            ),
            make_dummy_datasets_with_features(cardinality=[3, 10, 42]),
        ),
    ],
)
def test_deepstate_exceptions_with_feat_static_cat(estimator, datasets):
    with pytest.raises(Exception):
        estimator()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


import logging
from functools import partial
from pathlib import Path
from typing import Callable, Iterator, List, Optional, Type

import mxnet as mx
import numpy as np
from gluonts.core.serde import dump_json, load_json
from gluonts.dataset.common import DataEntry, Dataset
from gluonts.dataset.loader import DataBatch, InferenceDataLoader
from gluonts.model.forecast import Forecast
from gluonts.model.forecast_generator import (
    ForecastGenerator,
    SampleForecastGenerator,
    to_numpy,
)
from gluonts.model.predictor import OutputTransform, Predictor
from gluonts.mx.batchify import batchify
from gluonts.mx.component import equals
from gluonts.mx.context import get_mxnet_context
from gluonts.mx.util import (
    export_repr_block,
    export_symb_block,
    get_hybrid_forward_input_names,
    hybrid_block_to_symbol_block,
    import_repr_block,
    import_symb_block,
)
from gluonts.transform import Transformation


@to_numpy.register(mx.nd.NDArray)
def _(x: mx.nd.NDArray) -> np.ndarray:
    return x.asnumpy()


@to_numpy.register(mx.sym.Symbol)
def _(x: mx.sym.Symbol) -> np.ndarray:
    return x.asnumpy()


class GluonPredictor(Predictor):
    """
    Base predictor type for Gluon-based models.

    Parameters
    ----------
    input_names
        Input tensor names for the graph
    prediction_net
        Network that will be called for prediction
    batch_size
        Number of time series to predict in a single batch
    prediction_length
        Number of time steps to predict
    input_transform
        Input transformation pipeline
    output_transform
        Output transformation
    ctx
        MXNet context to use for computation
    forecast_generator
        Class to generate forecasts from network outputs
    """

    BlockType = mx.gluon.Block

    def __init__(
        self,
        input_names: List[str],
        prediction_net,
        batch_size: int,
        prediction_length: int,
        ctx: mx.Context,
        input_transform: Transformation,
        lead_time: int = 0,
        forecast_generator: ForecastGenerator = SampleForecastGenerator(),
        output_transform: Optional[OutputTransform] = None,
        dtype: Type = np.float32,
    ) -> None:
        super().__init__(
            lead_time=lead_time,
            prediction_length=prediction_length,
        )

        self.input_names = input_names
        self.prediction_net = prediction_net
        self.batch_size = batch_size
        self.input_transform = input_transform
        self.forecast_generator = forecast_generator
        self.output_transform = output_transform
        self.ctx = ctx
        self.dtype = dtype

    @property
    def network(self):
        return self.prediction_net

    def hybridize(self, batch: DataBatch) -> None:
        """
        Hybridizes the underlying prediction network.

        Parameters
        ----------
        batch
            A batch of data to use for the required forward pass after the
            `hybridize()` call.
        """
        self.prediction_net.hybridize(active=True)
        self.prediction_net(*[batch[k] for k in self.input_names])

    def as_symbol_block_predictor(
        self,
        batch: Optional[DataBatch] = None,
        dataset: Optional[Dataset] = None,
    ) -> "SymbolBlockPredictor":
        """
        Returns a variant of the current :class:`GluonPredictor` backed
        by a Gluon `SymbolBlock`. If the current predictor is already a
        :class:`SymbolBlockPredictor`, it just returns itself.

        One of batch or datset must be set.

        Parameters
        ----------
        batch
            A batch of data to use for the required forward pass after the
            `hybridize()` call of the underlying network.
        dataset
            Dataset from which a batch is extracted if batch is not set.

        Returns
        -------
        SymbolBlockPredictor
            A predictor derived from the current one backed by a `SymbolBlock`.
        """
        raise NotImplementedError

    def predict(
        self,
        dataset: Dataset,
        num_samples: Optional[int] = None,
        num_workers: Optional[int] = None,
        num_prefetch: Optional[int] = None,
        **kwargs,
    ) -> Iterator[Forecast]:
        inference_data_loader = InferenceDataLoader(
            dataset,
            transform=self.input_transform,
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.ctx, dtype=self.dtype),
        )
        with mx.Context(self.ctx):
            yield from self.forecast_generator(
                inference_data_loader=inference_data_loader,
                prediction_net=self.prediction_net,
                input_names=self.input_names,
                output_transform=self.output_transform,
                num_samples=num_samples,
            )

    def __eq__(self, that):
        if type(self) != type(that):
            return False

        if not equals(self.prediction_length, that.prediction_length):
            return False
        if not equals(self.lead_time, that.lead_time):
            return False
        if not equals(self.input_transform, that.input_transform):
            return False

        return equals(
            self.prediction_net.collect_params(),
            that.prediction_net.collect_params(),
        )

    def serialize(self, path: Path) -> None:
        # call Predictor.serialize() in order to serialize the class name
        super().serialize(path)

        # serialize every GluonPredictor-specific parameters
        # serialize the prediction network
        self.serialize_prediction_net(path)

        # serialize transformation chain
        with (path / "input_transform.json").open("w") as fp:
            print(dump_json(self.input_transform), file=fp)

        # FIXME: also needs to serialize the output_transform

        # serialize all remaining constructor parameters
        with (path / "parameters.json").open("w") as fp:
            parameters = dict(
                batch_size=self.batch_size,
                prediction_length=self.prediction_length,
                lead_time=self.lead_time,
                ctx=self.ctx,
                dtype=self.dtype,
                forecast_generator=self.forecast_generator,
                input_names=self.input_names,
            )
            print(dump_json(parameters), file=fp)

    def serialize_prediction_net(self, path: Path) -> None:
        raise NotImplementedError()


class SymbolBlockPredictor(GluonPredictor):
    """
    A predictor which serializes the network structure as an MXNet symbolic
    graph. Should be used for models deployed in production in order to ensure
    forward-compatibility as GluonTS models evolve.

    Used by the training shell if training is invoked with a hyperparameter
    `use_symbol_block_predictor = True`.
    """

    BlockType = mx.gluon.SymbolBlock

    def as_symbol_block_predictor(
        self,
        batch: Optional[DataBatch] = None,
        dataset: Optional[Dataset] = None,
    ) -> "SymbolBlockPredictor":
        return self

    def serialize_prediction_net(self, path: Path) -> None:
        export_symb_block(self.prediction_net, path, "prediction_net")

    @classmethod
    def deserialize(  # type: ignore
        cls, path: Path, ctx: Optional[mx.Context] = None
    ) -> "SymbolBlockPredictor":
        ctx = ctx if ctx is not None else get_mxnet_context()

        with mx.Context(ctx):
            # deserialize constructor parameters
            with (path / "parameters.json").open("r") as fp:
                parameters = load_json(fp.read())

            parameters["ctx"] = ctx

            # deserialize transformation chain
            with (path / "input_transform.json").open("r") as fp:
                transform = load_json(fp.read())

            # deserialize prediction network
            num_inputs = len(parameters["input_names"])
            prediction_net = import_symb_block(
                num_inputs, path, "prediction_net"
            )

            return SymbolBlockPredictor(
                input_transform=transform,
                prediction_net=prediction_net,
                **parameters,
            )


class RepresentableBlockPredictor(GluonPredictor):
    """
    A predictor which serializes the network structure using the JSON-
    serialization methods located in `gluonts.core.serde`. Use the following
    logic to create a `RepresentableBlockPredictor` from a trained prediction
    network.

    >>> def create_representable_block_predictor(
    ...        prediction_network: mx.gluon.HybridBlock,
    ...        **kwargs
    ... ) -> RepresentableBlockPredictor:
    ...    return RepresentableBlockPredictor(
    ...        prediction_net=prediction_network,
    ...        **kwargs
    ...    )
    """

    BlockType = mx.gluon.HybridBlock

    def __init__(
        self,
        prediction_net,
        batch_size: int,
        prediction_length: int,
        ctx: mx.Context,
        input_transform: Transformation,
        lead_time: int = 0,
        forecast_generator: ForecastGenerator = SampleForecastGenerator(),
        output_transform: Optional[
            Callable[[DataEntry, np.ndarray], np.ndarray]
        ] = None,
        dtype: Type = np.float32,
    ) -> None:
        super().__init__(
            input_names=get_hybrid_forward_input_names(type(prediction_net)),
            prediction_net=prediction_net,
            batch_size=batch_size,
            prediction_length=prediction_length,
            ctx=ctx,
            input_transform=input_transform,
            lead_time=lead_time,
            forecast_generator=forecast_generator,
            output_transform=output_transform,
            dtype=dtype,
        )

    def as_symbol_block_predictor(
        self,
        batch: Optional[DataBatch] = None,
        dataset: Optional[Dataset] = None,
    ) -> SymbolBlockPredictor:
        if batch is None:
            assert dataset is not None
            data_loader = InferenceDataLoader(
                dataset,
                transform=self.input_transform,
                batch_size=self.batch_size,
                stack_fn=partial(batchify, ctx=self.ctx, dtype=self.dtype),
            )
            batch = next(iter(data_loader))

        with self.ctx:
            symbol_block_net = hybrid_block_to_symbol_block(
                hb=self.prediction_net,
                data_batch=[batch[k] for k in self.input_names],
            )

        return SymbolBlockPredictor(
            input_names=self.input_names,
            prediction_net=symbol_block_net,
            batch_size=self.batch_size,
            prediction_length=self.prediction_length,
            ctx=self.ctx,
            input_transform=self.input_transform,
            lead_time=self.lead_time,
            forecast_generator=self.forecast_generator,
            output_transform=self.output_transform,
            dtype=self.dtype,
        )

    def serialize(self, path: Path) -> None:
        logging.warning(
            "Serializing RepresentableBlockPredictor instances does not save "
            "the prediction network structure in a backwards-compatible "
            "manner. Be careful not to use this method in production."
        )
        super().serialize(path)

    def serialize_prediction_net(self, path: Path) -> None:
        export_repr_block(self.prediction_net, path, "prediction_net")

    @classmethod
    def deserialize(  # type: ignore
        cls, path: Path, ctx: Optional[mx.Context] = None
    ) -> "RepresentableBlockPredictor":
        ctx = ctx if ctx is not None else get_mxnet_context()

        with mx.Context(ctx):
            # deserialize constructor parameters
            with (path / "parameters.json").open("r") as fp:
                parameters = load_json(fp.read())

            # deserialize transformation chain
            with (path / "input_transform.json").open("r") as fp:
                transform = load_json(fp.read())

            # deserialize prediction network
            prediction_net = import_repr_block(path, "prediction_net")

            # input_names is derived from the prediction_net
            if "input_names" in parameters:
                del parameters["input_names"]

            parameters["ctx"] = ctx

            return RepresentableBlockPredictor(
                input_transform=transform,
                prediction_net=prediction_net,
                **parameters,
            )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pytest

from mxnet import nd

from gluonts.mx.model.gp_forecaster.gaussian_process import GaussianProcess

from gluonts.mx.kernels import RBFKernel

from .data import (
    load_exact_mean,
    load_exact_std,
    load_gp_params,
    load_xfull,
    load_ytrain,
)


def relative_error(y_hat, y_exact):
    return nd.max(
        nd.max(nd.abs(y_exact - y_hat), axis=1)
        / nd.max(nd.abs(y_exact), axis=1)
    )


@pytest.mark.parametrize(
    "gp_params, mean_exact, std_exact, x_full, y_train",
    [
        # Test inference when training and test length are equal
        (
            load_gp_params(),
            load_exact_mean(),
            load_exact_std(),
            load_xfull(),
            load_ytrain(),
        ),
        # Test inference when training and test length are not equal
        (
            load_gp_params(),
            load_exact_mean()[:, :72],
            load_exact_std()[:, :72],
            load_xfull(),
            load_ytrain(),
        ),
    ],
)
def test_inference(gp_params, mean_exact, std_exact, x_full, y_train) -> None:
    # Initialize problem parameters
    tol = 1e-2
    num_samples = 100
    context_length = y_train.shape[1]
    prediction_length = mean_exact.shape[1]

    # Extract training and test set
    x_train = x_full[:, :context_length, :]
    x_test = x_full[:, context_length : context_length + prediction_length, :]

    amplitude = gp_params[:, 0, :].expand_dims(axis=2)
    length_scale = gp_params[:, 1, :].expand_dims(axis=2)
    sigma = gp_params[:, 2, :].expand_dims(axis=2)

    # Instantiate RBFKernel with its hyper-parameters
    kernel = RBFKernel(amplitude, length_scale)

    # Instantiate gp_inference object and hybridize
    gp = GaussianProcess(
        sigma=sigma,
        kernel=kernel,
        context_length=context_length,
        prediction_length=prediction_length,
        num_samples=num_samples,
        float_type=np.float32,
    )

    # Compute predictive mean and covariance
    _, mean, std = gp.exact_inference(x_train, y_train, x_test)

    # This test compares to the predictive mean and std generated from MATLAB's fitrgp, which
    # outputs the sample with the noise, i.e. adds :math:`sigma^2` to the diagonal of
    # the predictive covariance matrix.
    assert relative_error(mean, mean_exact) <= tol
    assert relative_error(std, std_exact) <= tol



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from typing import List

from mxnet.gluon import HybridBlock, nn

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.loader import (
    DataLoader,
    TrainDataLoader,
    ValidationDataLoader,
)
from gluonts.model.predictor import Predictor
from gluonts.mx.batchify import batchify
from gluonts.mx.block.feature import FeatureEmbedder
from gluonts.mx.block.rnn import RNN
from gluonts.mx.distribution import DistributionOutput, StudentTOutput
from gluonts.mx.model.estimator import GluonEstimator
from gluonts.mx.model.predictor import RepresentableBlockPredictor
from gluonts.mx.trainer import Trainer
from gluonts.mx.util import get_hybrid_forward_input_names
from gluonts.time_feature import time_features_from_frequency_str
from gluonts.transform import (
    AddTimeFeatures,
    AsNumpyArray,
    InstanceSplitter,
    SelectFields,
    SetFieldIfNotPresent,
    TestSplitSampler,
    Transformation,
    ValidationSplitSampler,
)

from ._network import CanonicalPredictionNetwork, CanonicalTrainingNetwork


class CanonicalEstimator(GluonEstimator):
    @validated()
    def __init__(
        self,
        model: HybridBlock,
        is_sequential: bool,
        freq: str,
        context_length: int,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        num_parallel_samples: int = 100,
        cardinality: List[int] = list([1]),
        embedding_dimension: int = 10,
        distr_output: DistributionOutput = StudentTOutput(),
        batch_size: int = 32,
    ) -> None:
        super().__init__(trainer=trainer, batch_size=batch_size)

        # TODO: error checking
        self.freq = freq
        self.context_length = context_length
        self.prediction_length = prediction_length
        self.distr_output = distr_output
        self.num_parallel_samples = num_parallel_samples
        self.cardinality = cardinality
        self.embedding_dimensions = [embedding_dimension for _ in cardinality]
        self.model = model
        self.is_sequential = is_sequential

    def create_transformation(self) -> Transformation:
        return (
            AsNumpyArray(field=FieldName.TARGET, expected_ndim=1)
            + AddTimeFeatures(
                start_field=FieldName.START,
                target_field=FieldName.TARGET,
                output_field=FieldName.FEAT_TIME,
                time_features=time_features_from_frequency_str(self.freq),
                pred_length=self.prediction_length,
            )
            + SetFieldIfNotPresent(
                field=FieldName.FEAT_STATIC_CAT, value=[0.0]
            )
            + AsNumpyArray(field=FieldName.FEAT_STATIC_CAT, expected_ndim=1)
        )

    def _create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        instance_sampler = {
            "training": ValidationSplitSampler(
                min_future=self.prediction_length
            ),
            "validation": ValidationSplitSampler(
                min_future=self.prediction_length
            ),
            "test": TestSplitSampler(),
        }[mode]

        return InstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=instance_sampler,
            time_series_fields=[FieldName.FEAT_TIME],
            past_length=self.context_length,
            future_length=self.prediction_length,
        )

    def create_training_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(CanonicalTrainingNetwork)
        instance_splitter = self._create_instance_splitter("training")
        return TrainDataLoader(
            dataset=data,
            transform=instance_splitter + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.trainer.ctx, dtype=self.dtype),
            **kwargs,
        )

    def create_validation_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(CanonicalTrainingNetwork)
        instance_splitter = self._create_instance_splitter("validation")
        return ValidationDataLoader(
            dataset=data,
            transform=instance_splitter + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.trainer.ctx, dtype=self.dtype),
        )

    def create_training_network(self) -> CanonicalTrainingNetwork:
        return CanonicalTrainingNetwork(
            embedder=FeatureEmbedder(
                cardinalities=self.cardinality,
                embedding_dims=self.embedding_dimensions,
            ),
            model=self.model,
            distr_output=self.distr_output,
            is_sequential=self.is_sequential,
        )

    def create_predictor(
        self,
        transformation: Transformation,
        trained_network: CanonicalTrainingNetwork,
    ) -> Predictor:
        prediction_splitter = self._create_instance_splitter("test")

        prediction_net = CanonicalPredictionNetwork(
            embedder=trained_network.embedder,
            model=trained_network.model,
            distr_output=trained_network.distr_output,
            is_sequential=trained_network.is_sequential,
            prediction_len=self.prediction_length,
            num_parallel_samples=self.num_parallel_samples,
            params=trained_network.collect_params(),
        )

        return RepresentableBlockPredictor(
            input_transform=transformation + prediction_splitter,
            prediction_net=prediction_net,
            batch_size=self.batch_size,
            prediction_length=self.prediction_length,
            ctx=self.trainer.ctx,
        )


class CanonicalRNNEstimator(CanonicalEstimator):
    @validated()
    def __init__(
        self,
        freq: str,
        context_length: int,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        num_layers: int = 1,
        num_cells: int = 50,
        cell_type: str = "lstm",
        num_parallel_samples: int = 100,
        cardinality: List[int] = list([1]),
        embedding_dimension: int = 10,
        distr_output: DistributionOutput = StudentTOutput(),
    ) -> None:
        model = RNN(
            mode=cell_type, num_layers=num_layers, num_hidden=num_cells
        )

        super().__init__(
            model=model,
            is_sequential=True,
            freq=freq,
            context_length=context_length,
            prediction_length=prediction_length,
            trainer=trainer,
            num_parallel_samples=num_parallel_samples,
            cardinality=cardinality,
            embedding_dimension=embedding_dimension,
            distr_output=distr_output,
        )


class MLPForecasterEstimator(CanonicalEstimator):
    @validated()
    def __init__(
        self,
        freq: str,
        context_length: int,
        prediction_length: int,
        trainer: Trainer = Trainer(),
        hidden_dim_sequence=list([50]),
        num_parallel_samples: int = 100,
        cardinality: List[int] = list([1]),
        embedding_dimension: int = 10,
        distr_output: DistributionOutput = StudentTOutput(),
    ) -> None:
        model = nn.HybridSequential()

        for layer, layer_dim in enumerate(hidden_dim_sequence):
            model.add(
                nn.Dense(
                    layer_dim,
                    flatten=False,
                    activation="relu",
                    prefix="mlp_%d_" % layer,
                )
            )

        super().__init__(
            model=model,
            is_sequential=False,
            freq=freq,
            context_length=context_length,
            prediction_length=prediction_length,
            trainer=trainer,
            num_parallel_samples=num_parallel_samples,
            cardinality=cardinality,
            embedding_dimension=embedding_dimension,
            distr_output=distr_output,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pandas as pd
import pytest

from gluonts.evaluation.backtest import make_evaluation_predictions
from gluonts.model.trivial.identity import IdentityPredictor


@pytest.mark.parametrize("freq", ["23min", "H", "17H", "9D", "3W", "5M"])
def test_make_evaluation_predictions_zero(freq):
    prediction_length = 5
    ts_length = 50

    predictor = IdentityPredictor(
        prediction_length=prediction_length, num_samples=5
    )
    dataset = [
        dict(
            start=pd.Period("2010-01-01 00:00", freq=freq) + k,
            target=np.arange(start=k, stop=k + ts_length),
        )
        for k in range(10)
    ]

    forecasts, tss = make_evaluation_predictions(dataset, predictor)

    for k, (forecast, ts) in enumerate(zip(forecasts, tss)):
        assert np.allclose(
            forecast.median,
            np.arange(
                start=ts_length - 2 * prediction_length + k,
                stop=ts_length - prediction_length + k,
            ),
        )
        assert isinstance(ts, pd.DataFrame)
        assert isinstance(ts.index, pd.PeriodIndex)
        assert ts.index[0] == dataset[k]["start"]
        assert ts.shape[0] == len(dataset[k]["target"])
        assert np.allclose(
            ts.iloc[:, 0].values, np.arange(start=k, stop=k + ts_length)
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np

import pytest

from gluonts import transform
from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.mx.model.seq2seq._transform import ForkingSequenceSplitter
from gluonts.time_feature import time_features_from_frequency_str

# if we import TestSplitSampler as Test... pytest thinks it's a test
from gluonts.transform import TestSplitSampler as TSplitSampler
from gluonts.transform import ValidationSplitSampler


def make_dataset(N, train_length):
    # generates 2 ** N - 1 timeseries with constant increasing values
    n = 2**N - 1

    targets = np.arange(n * train_length).reshape((n, train_length))

    return ListDataset(
        [{"start": "2012-01-01", "target": targets[i, :]} for i in range(n)],
        freq="D",
    )


def test_forking_sequence_splitter() -> None:
    len_ts = 20
    ds = make_dataset(1, len_ts)
    enc_len = 5
    dec_len = 3

    trans = transform.Chain(
        [
            transform.AddAgeFeature(
                target_field=FieldName.TARGET,
                output_field="age",
                pred_length=dec_len,
            ),
            ForkingSequenceSplitter(
                instance_sampler=ValidationSplitSampler(min_future=dec_len),
                enc_len=enc_len,
                dec_len=dec_len,
                encoder_series_fields=["age"],
            ),
        ]
    )

    out = trans(ds, is_train=True)
    transformed_data = next(iter(out))

    future_target = np.array(
        [
            [13.0, 14.0, 15.0],
            [14.0, 15.0, 16.0],
            [15.0, 16.0, 17.0],
            [16.0, 17.0, 18.0],
            [17.0, 18.0, 19.0],
        ]
    )
    assert (
        np.linalg.norm(future_target - transformed_data["future_target"])
        < 1e-5
    ), "the forking sequence target should be computed correctly."

    age = np.log10(2.0 + np.arange(len_ts))
    assert (
        np.linalg.norm(
            age[-(enc_len + dec_len) : -dec_len]
            - transformed_data["past_age"].flatten()
        )
        < 1e-5
    ), "the forking sequence past feature should be computed correctly."


@pytest.mark.parametrize("is_train", [True, False])
def test_forking_sequence_with_features(is_train) -> None:
    def make_dataset(N, train_length):
        # generates 2 ** N - 1 timeseries with constant increasing values
        n = 2**N - 1

        targets = np.arange(n * train_length).reshape((n, train_length))

        return ListDataset(
            [
                {"start": "2012-01-01", "target": targets[i, :]}
                for i in range(n)
            ],
            freq="D",
        )

    ds = make_dataset(1, 20)
    enc_len = 5
    dec_len = 3
    num_forking = 1
    num_time_feat_daily_freq = 3
    num_age_feat = 1

    trans = transform.Chain(
        [
            transform.AddAgeFeature(
                target_field=FieldName.TARGET,
                output_field=FieldName.FEAT_AGE,
                pred_length=10,
            ),
            transform.AddTimeFeatures(
                start_field=FieldName.START,
                target_field=FieldName.TARGET,
                output_field=FieldName.FEAT_TIME,
                time_features=time_features_from_frequency_str("D"),
                pred_length=10,
            ),
            ForkingSequenceSplitter(
                instance_sampler=(
                    ValidationSplitSampler(min_future=dec_len)
                    if is_train
                    else TSplitSampler()
                ),
                enc_len=enc_len,
                dec_len=dec_len,
                num_forking=num_forking,
                encoder_series_fields=[
                    FieldName.FEAT_AGE,
                    FieldName.FEAT_TIME,
                ],
                decoder_series_fields=[FieldName.FEAT_TIME],
            ),
        ]
    )

    out = trans(iter(ds), is_train=is_train)
    transformed_data = next(iter(out))

    assert transformed_data["past_target"].shape == (enc_len, 1)
    assert transformed_data["past_feat_dynamic_age"].shape == (
        enc_len,
        num_age_feat,
    )
    assert transformed_data["past_time_feat"].shape == (
        enc_len,
        num_time_feat_daily_freq,
    )
    assert transformed_data["future_time_feat"].shape == (
        num_forking,
        dec_len,
        num_time_feat_daily_freq,
    )

    if is_train:
        assert transformed_data["future_target"].shape == (
            num_forking,
            dec_len,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
``zebras.Period`` and ``zebras.Periods`` are classes to track points in time
with periodicity. They provide similar functionality to ``pandas.Period`` and
``pandas.PeriodIndex`` but offer some advantages:

* more consistent API
* improved ergonomics
* better performance

Both classes are just thin layers around ``numpy.datetime64`` objects, but
provide a more usable interface.

Both classes have easier to use factory-functions in ``zebras.period`` and
``zebras.periods``, akin to how ``numpy.ndarray`` objects are usually
constructed using ``numpy.array``.

While ``zebras.Period`` represents a single time-stamp, ``zebras.Periods`` are
a set of equidistant timestamps where the gap between consecutive timestamps
is the period.


``py
first = zb.period("2020-01", "3M")
index = zb.periods("2020-01", "3M", 3)

first.periods(3) == index
first == index[0]
```
"""

from __future__ import annotations

import datetime
import functools
from dataclasses import dataclass
from typing import Any, Union, Optional, Tuple, cast, overload

import numpy as np
from dateutil.parser import parse as du_parse

from gluonts.core import serde

from ._freq import Freq


def _is_number(value):
    return isinstance(value, (int, np.integer))


class _BasePeriod:
    data: Any
    freq: Freq

    @property
    def freqstr(self) -> str:
        return str(self.freq)

    @property
    def year(self) -> np.ndarray:
        return self.data.astype("M8[Y]").astype(int) + 1970

    @property
    def month(self) -> np.ndarray:
        return self.data.astype("M8[M]").astype(int) % 12 + 1

    @property
    def day(self) -> np.ndarray:
        return (self.data.astype("M8[D]") - self.data.astype("M8[M]")).astype(
            int
        ) + 1

    @property
    def hour(self) -> np.ndarray:
        return (self.data.astype("M8[h]") - self.data.astype("M8[D]")).astype(
            int
        )

    @property
    def minute(self) -> np.ndarray:
        return (self.data.astype("M8[m]") - self.data.astype("M8[h]")).astype(
            int
        )

    @property
    def second(self) -> np.ndarray:
        return (self.data.astype("M8[s]") - self.data.astype("M8[m]")).astype(
            int
        )

    @property
    def dayofweek(self) -> np.ndarray:
        return (self.data.astype("M8[D]").astype(int) - 4) % 7

    @property
    def dayofyear(self) -> np.ndarray:
        return (self.data.astype("M8[D]") - self.data.astype("M8[Y]")).astype(
            int
        ) + 1

    @property
    def week(self) -> np.ndarray:
        # Note: In Python 3.9 `isocalendar()` returns a named tuple, but we
        # need to support 3.7 and 3.8, so we use index one for the week.
        return np.array(
            [
                cal.isocalendar()[1]
                for cal in self.data.astype(datetime.datetime)
            ]
        )

    def __add__(self, other):
        if _is_number(other):
            return self.__class__(
                self.freq.shift(self.data, other),
                self.freq,
            )

    def __sub__(self, other):
        if _is_number(other):
            return self.__class__(
                self.freq.shift(self.data, -other),
                self.freq,
            )

        else:
            return self.data - other.data

        raise ValueError(other)


@functools.total_ordering
@dataclass
class Period(_BasePeriod):
    data: np.datetime64
    freq: Freq

    @property
    def __init_passed_kwargs__(self) -> dict:
        return {"data": self.data, "freq": self.freq}

    def periods(self, count: int):
        return Periods(
            self.freq.range(self.data, count),
            self.freq,
        )

    def to_pandas(self):
        import pandas as pd

        return pd.Period(self.data.astype(object), self.freq.to_pandas())

    def to_timestamp(self):
        return self.data.astype(object)

    def unix_epoch(self) -> int:
        return self.to_numpy().astype("M8[s]").astype(int)

    def __repr__(self) -> str:
        return f"Period<{self.data}, {self.freq}>"

    def __lt__(self, other: Period) -> bool:
        # convert numpy.bool_ into bool
        return cast(bool, self.data < other.data)

    def to_numpy(self) -> np.datetime64:
        return self.data

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.data.shape

    def __array__(self) -> np.datetime64:
        return self.data


@dataclass
class Periods(_BasePeriod):
    data: np.ndarray
    freq: Freq

    @property
    def start(self) -> Period:
        return self[0]

    @property
    def end(self) -> Period:
        """
        Last timestamp.

        >>> p = periods("2021", "D", 365)
        >>> assert p.end == period("2021-12-31", "D")
        """

        return self[-1]

    def head(self, count: int) -> Periods:
        """
        First ``count`` timestamps.

        >>> p = periods("2021", "D", 365)
        >>> assert p.head(5) == periods("2021-01-01", "D", 5)
        """

        return self[:count]

    def tail(self, count: int) -> Periods:
        """
        Last ``count`` timestamps.

        >>> p = periods("2021", "D", 365)
        >>> assert p.tail(5) == periods("2021-12-27", "D", 5)
        """

        return self[-count:]

    def future(self, count: int) -> Periods:
        """
        Next ``count`` timestamps.

        >>> p = periods("2021", "D", 365)
        >>> assert p.future(5) == periods("2022-01-01", "D", 5)
        """
        return (self.end + 1).periods(count)

    def past(self, count: int) -> Periods:
        """
        Previous ``count`` timestamps.

        >>> p = periods("2021", "D", 365)
        >>> assert p.past(5) == periods("2020-12-27", "D", 5)
        """

        return (self.start - count).periods(count)

    def prepend(self, count: int) -> Periods:
        """
        Copy which contains ``count`` past timestamps.

        >>> p = periods("2021", "D", 365)
        >>> assert p.prepend(5) == periods("2020-12-27", "D", 370)
        """
        return Periods(
            np.concatenate([self.past(count).data, self.data]),
            self.freq,
        )

    def extend(self, count: int) -> Periods:
        """
        Copy which contains ``count`` future timestamps.

        >>> p = periods("2021", "D", 365)
        >>> assert p.extend(5) == periods("2021", "D", 370)
        """
        return Periods(
            np.concatenate([self.data, self.future(count).data]),
            self.freq,
        )

    def to_pandas(self):
        import pandas as pd

        # older versions of pandas expect ns-datetime64
        return pd.PeriodIndex(
            self.data.astype("M8[ns]"), freq=self.freq.to_pandas()
        )

    @classmethod
    def from_pandas(cls, index):
        """
        Turn ``pandas.PeriodIndex`` or ``pandas.DatetimeIndex`` into
        ``Periods``.
        """

        import pandas as pd

        if isinstance(index, pd.DatetimeIndex):
            index = index.to_period()
        else:
            assert isinstance(index, pd.PeriodIndex)

        freq = Freq.from_pandas(index.freqstr)
        np_index = np.array(index.asi8, dtype=f"M8[{freq.np_freq[0]}]")
        assert np.all(np.diff(np_index).astype(int) == freq.n)

        return Periods(np_index, freq)

    def intersection(self, other):
        # TODO: Is this needed?
        return self.data[np.in1d(self, other)]

    def index_of(self, period: Union[str, Period]):
        """
        Return the index of ``period``

        >>> p = periods("2021", "D", 365)
        >>> assert p.index_of(period("2021-02-01", "D")) == 31
        """

        if isinstance(period, str):
            period = Period(
                np.datetime64(du_parse(period), self.freq.np_freq), self.freq
            )

        idx = (period - self.start).astype(int) // self.freq.step
        assert 0 <= idx <= len(self), idx

        return idx

    def __len__(self):
        return len(self.data)

    @overload
    def __getitem__(self, idx: int) -> Period: ...

    @overload
    def __getitem__(self, idx: slice) -> Periods: ...

    def __getitem__(self, idx):
        if _is_number(idx):
            return Period(self.data[idx], self.freq)

        return Periods(self.data[idx], self.freq)

    def __eq__(self, other):
        if not isinstance(other, Periods):
            return False

        return len(self) == len(other) and self.start == other.start

    def to_numpy(self) -> np.ndarray:
        return self.data

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.data.shape

    def __array__(self) -> np.ndarray:
        return self.data

    def unix_epoch(self) -> np.ndarray:
        return self.to_numpy().astype("M8[s]").astype(int)


@serde.encode.register
def _encode_zebras_periods(v: Periods):
    return {
        "__kind__": "instance",
        "class": "gluonts.zebras.periods",
        "kwargs": serde.encode(
            {"start": v.start, "freq": str(v.freq), "count": len(v)}
        ),
    }


def period(
    data: Union[Period, str], freq: Optional[Union[Freq, str]] = None
) -> Period:
    """
    Create a ``zebras.Period`` object that represents a period of time.

    Parameters
    ----------
    data
        The time period represented by a string (e.g., "2023-01-01"),
        or another Period object.
    freq, optional
        The frequency of the period, e.g, "H" for hourly, by default None.

    Returns
    -------
        A ``zebras.Period`` object.
    """
    if freq is None:
        if hasattr(data, "freqstr"):
            freq = Freq.from_pandas(data.freqstr)
        else:
            raise ValueError("No frequency specified.")
    elif isinstance(freq, Freq):
        freq = freq
    elif isinstance(freq, str):
        freq = Freq.from_pandas(freq)
    else:
        raise ValueError(f"Unknown frequency type {type(freq)}.")

    data_: Any

    if isinstance(data, Period):
        data_ = data.data

    elif isinstance(data, str):
        data_ = du_parse(
            data,
            default=datetime.datetime(1970, 1, 1),
            ignoretz=True,
        )
    else:
        # TODO: should we add a check?
        data_ = data

    return Period(freq.align(np.datetime64(data_, freq.np_freq)), freq)


def periods(
    start: Union[Period, str], freq: Union[Freq, str], count: int
) -> Period:
    """
    Create a ``zebras.Periods`` object that represents multiple consecutive
    periods of time.

    Parameters
    ----------
    start
        The starting time period represented by a string (e.g., "2023-01-01"),
        or another Period object.
    freq
        The frequency of the period, e.g, "H" for hourly.
    count
        The number of periods.

    Returns
    -------
        A ``zebras.Periods`` object.
    """
    return period(start, freq).periods(count)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx

import numpy as np

from gluonts.mx.distribution.gaussian import Gaussian
from gluonts.mx.distribution.mixture import MixtureDistribution
from gluonts.mx.distribution.uniform import Uniform

# In these tests we can use only distributions that have a cdf method but do not have a quantile method.
# The mixture distribution satisfies these conditions. The components of the mixture should have a cdf method.


def test_quantile() -> None:
    r"""
    Tests if the quantiles of a single Gaussian and the quantiles of the
    mixture of two Gaussians identical to the first are equal.

    The quantiles of the single Gaussian are given by the Gaussian.quantile()
    method while the quantiles of the mixture from the Distribution.quantile()
    method.
    """
    mu = mx.nd.array(
        [[1, 10, 100, 1000, 10000], [-1, -10, -100, -1000, -10000]]
    )
    sigma = mx.nd.array([[1.0, 2.0, 3.0, 4.0, 5.0]] * 2)

    gau = Gaussian(mu, sigma)

    mixture_probs = mx.nd.broadcast_like(
        mx.nd.array([[0.5, 0.5]]).expand_dims(axis=0),
        mu,
        lhs_axes=(0, 1),
        rhs_axes=(0, 1),
    )
    mix = MixtureDistribution(
        mixture_probs=mixture_probs,
        components=[Gaussian(mu, sigma), Gaussian(mu, sigma)],
    )

    quantiles = mx.nd.array([0.1, 0.5, 0.9])
    gau_quantiles = gau.quantile(quantiles)
    mix_quantiles = mix.quantile(quantiles)

    relative_error = np.max(
        np.abs(gau_quantiles.asnumpy() / mix_quantiles.asnumpy() - 1)
    )
    assert relative_error < 1e-4


def test_inverse_quantile() -> None:
    r"""
    Tests CDF(Q(x)) == x.
    """
    mu = mx.nd.array([1])
    sigma = mx.nd.array([0.1])
    uni_bound = mx.nd.array([1])
    mixture_probs = mx.nd.array([[0.3, 0.7]])

    levels = mx.nd.array([0.01, 0.1, 0.5, 0.9, 0.99])

    mix_gau = MixtureDistribution(
        mixture_probs=mixture_probs,
        components=[Gaussian(mu, sigma), Gaussian(mu, sigma)],
    )
    mix_uni = MixtureDistribution(
        mixture_probs=mixture_probs,
        components=[
            Uniform(uni_bound, 3 * uni_bound),
            Uniform(uni_bound, 3 * uni_bound),
        ],
    )
    mix_gau_uni = MixtureDistribution(
        mixture_probs=mixture_probs,
        components=[Gaussian(mu, sigma), Uniform(uni_bound, 3 * uni_bound)],
    )

    gau_res = mix_gau.cdf(mix_gau.quantile(levels))
    uni_res = mix_uni.cdf(mix_uni.quantile(levels))
    gau_uni_res = mix_gau_uni.cdf(mix_gau_uni.quantile(levels))

    assert (
        np.max(gau_res.asnumpy() / levels.asnumpy().reshape(-1, 1) - 1) < 1e-4
    )
    assert (
        np.max(uni_res.asnumpy() / levels.asnumpy().reshape(-1, 1) - 1) < 1e-4
    )
    assert (
        np.max(gau_uni_res.asnumpy() / levels.asnumpy().reshape(-1, 1) - 1)
        < 1e-4
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import lightning.pytorch as pl
import torch

from gluonts.core.component import validated

from .module import WaveNet


class WaveNetLightningModule(pl.LightningModule):
    """
    LightningModule wrapper over WaveNet.

    Parameters
    ----------
    model_kwargs
        Keyword arguments to pass to WaveNet.
    lr, optional
        Learning rate, by default 1e-3
    weight_decay, optional
        Weight decay, by default 1e-8
    """

    @validated()
    def __init__(
        self,
        model_kwargs: dict,
        lr: float = 1e-3,
        weight_decay: float = 1e-8,
    ) -> None:
        super().__init__()
        self.save_hyperparameters()
        self.model = WaveNet(**model_kwargs)
        self.lr = lr
        self.weight_decay = weight_decay

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def training_step(self, batch, batch_idx: int):  # type: ignore
        """
        Execute training step.
        """
        feat_static_cat = batch["feat_static_cat"]
        feat_static_real = batch["feat_static_real"]
        past_target = batch["past_target"]
        past_observed_values = batch["past_observed_values"]
        past_time_feat = batch["past_time_feat"]
        future_target = batch["future_target"]
        future_observed_values = batch["future_observed_values"]
        future_time_feat = batch["future_time_feat"]
        scale = batch["scale"]

        train_loss = self.model.loss(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_target=past_target,
            past_observed_values=past_observed_values,
            past_time_feat=past_time_feat,
            future_target=future_target,
            future_observed_values=future_observed_values,
            future_time_feat=future_time_feat,
            scale=scale,
        ).mean()

        self.log(
            "train_loss",
            train_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )

        return train_loss

    def validation_step(self, batch, batch_idx: int):  # type: ignore
        """
        Execute validation step.
        """
        feat_static_cat = batch["feat_static_cat"]
        feat_static_real = batch["feat_static_real"]
        past_target = batch["past_target"]
        past_observed_values = batch["past_observed_values"]
        past_time_feat = batch["past_time_feat"]
        future_target = batch["future_target"]
        future_observed_values = batch["future_observed_values"]
        future_time_feat = batch["future_time_feat"]
        scale = batch["scale"]

        val_loss = self.model.loss(
            feat_static_cat=feat_static_cat,
            feat_static_real=feat_static_real,
            past_target=past_target,
            past_observed_values=past_observed_values,
            past_time_feat=past_time_feat,
            future_target=future_target,
            future_observed_values=future_observed_values,
            future_time_feat=future_time_feat,
            scale=scale,
        ).mean()

        self.log(
            "val_loss",
            val_loss,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )

        return val_loss

    def configure_optimizers(self):
        """
        Returns the optimizer to use.
        """
        return torch.optim.AdamW(
            self.model.parameters(),
            lr=self.lr,
            weight_decay=self.weight_decay,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path

import pandas as pd
from gluonts.dataset import DatasetWriter
from gluonts.dataset.common import MetaData, TrainDatasets


# Currently data from only two regions are made public.
NUM_REGIONS = 2


def generate_ett_small_dataset(
    dataset_path: Path,
    dataset_writer: DatasetWriter,
    base_file_name: str,
    freq: str,
    prediction_length: int,
):
    dfs = []
    for i in range(NUM_REGIONS):
        df = pd.read_csv(
            f"https://raw.githubusercontent.com/zhouhaoyi/ETDataset"
            f"/main/ETT-small/{base_file_name}{i+1}.csv"
        )
        df["date"] = df["date"].astype("datetime64[ms]")
        dfs.append(df)

    test = []
    for region, df in enumerate(dfs):
        start = pd.Period(df["date"][0], freq=freq)
        for col in df.columns:
            if col in ["date"]:
                continue
            test.append(
                {
                    "start": start,
                    "item_id": f"{col}_{region}",
                    "target": df[col].values,
                }
            )

    train = []
    for region, df in enumerate(dfs):
        start = pd.Period(df["date"][0], freq=freq)
        for col in df.columns:
            if col in ["date"]:
                continue
            train.append(
                {
                    "start": start,
                    "item_id": f"{col}_{region}",
                    "target": df[col].values[:-prediction_length],
                }
            )

    metadata = MetaData(freq=freq, prediction_length=prediction_length)
    dataset = TrainDatasets(metadata=metadata, train=train, test=test)
    dataset.save(str(dataset_path), writer=dataset_writer, overwrite=True)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from gluonts.core.settings import Settings, let
from gluonts.pydantic import BaseModel


class MySettings(Settings):
    foo: str = "bar"


class Args(BaseModel):
    a: int = 1
    b: int = 2


settings = MySettings()
settings._declare("args", Args, default=Args())


def test_functional():
    settings = Settings()
    settings._declare("foo", str, default="bar")

    assert settings.foo == "bar"


def test_declarative():
    assert settings.foo == "bar"


def test_get():
    with settings._let(foo="hello"):
        assert settings["foo"] == "hello"
        assert settings.foo == "hello"
        assert settings._get("foo") == "hello"


def test_let():
    assert settings.foo == "bar"

    with let(settings, foo="hello"):
        assert settings.foo == "hello"

    assert settings.foo == "bar"

    with settings._let(foo="hello"):
        assert settings.foo == "hello"

    assert settings.foo == "bar"


def test_inject():
    @settings._inject("a")
    def x(a, b):
        return a, b

    with settings._let(a=42):
        assert x(b=0) == (42, 0)
        assert x(1, 2) == (1, 2)

    @settings._inject("b")
    def x(a, b):
        return a, b

    with settings._let(b=42):
        assert x(0) == (0, 42)
        assert x(1, 2) == (1, 2)

    @settings._inject("a")
    def x(a=1):
        return a

    assert x() == 1

    @settings._inject(a="args.a", b="args.b")
    def x(a, b):
        return a, b

    assert x() == (1, 2)

    with settings._let(args={"a": 4, "b": 5}):
        assert x() == (4, 5)


def test_partial_assignment():
    assert settings.args.a == 1

    settings._set("args", {"a": 9})
    assert settings.args.a == 9

    settings._set("args", {"b": 42})
    assert settings.args.b == 42

    with settings._let(args=dict(b=3)):
        assert settings.args.a == 9
        assert settings.args.b == 3

    settings._set("args", {"a": "1"})
    assert settings.args.a == 1



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import field
from typing import Dict, Any, Optional

from typing_extensions import Literal

import numpy as np
import mxnet as mx
from mxnet import gluon
import mxnet.gluon.nn as nn

from gluonts.core.component import validated
from gluonts.pydantic import dataclass

from .callback import Callback


@dataclass
class Objective:
    best: float

    @staticmethod
    def from_str(s: Literal["min", "max"]) -> "Objective":
        if s == "min":
            return Min()
        else:
            return Max()

    def update(self, metric: float) -> bool:
        if self.should_update(metric):
            self.best = metric
            return True
        return False

    def should_update(self, metric: float) -> bool:
        raise NotImplementedError


@dataclass
class Min(Objective):
    best: float = np.Inf

    def should_update(self, metric: float) -> bool:
        return metric < self.best


@dataclass
class Max(Objective):
    best: float = -np.Inf

    def should_update(self, metric: float) -> bool:
        return metric > self.best


@dataclass
class Patience:
    """
    Simple patience tracker.

    Given an `Objective`, it will check whether the metric has improved and
    update its patience count. A better value sets the patience back to zero.

    In addition, one needs to call ``reset()`` explicitly after the patience
    was exceeded, otherwise `RuntimError` is raised when trying to invoke
    `step`.

    ``Patience`` keeps track of the number of invocations to ``reset``, via
    ``num_resets``.
    """

    patience: int = field(metadata={"ge": 0})
    objective: Objective
    current_patience: int = field(default=0, init=False)
    num_resets: int = field(default=0, init=False)
    exceeded: bool = field(default=False, init=False)

    def reset(self) -> None:
        self.current_patience = 0
        self.exceeded = False
        self.num_resets += 1

    def step(self, metric_value: float) -> bool:
        if self.exceeded:
            raise RuntimeError("Patience already exceeded.")

        has_improved = self.objective.update(metric_value)

        if has_improved:
            self.current_patience = 0
        else:
            self.current_patience += 1

        # this can also trigger in case of improvement when `self.patience = 0`
        self.exceeded = self.current_patience >= self.patience
        return self.exceeded


@dataclass
class MetricAttentiveScheduler:
    """
    This scheduler decreases the learning rate based on the value of some
    validation metric to be optimized (maximized or minimized). The value of
    such metric is provided by calling the `step` method on the scheduler. A
    `patience` parameter must be provided, and the scheduler will reduce the
    learning rate if no improvement in the metric is done before `patience`
    observations of the metric.

    Examples:

        `patience = 0`: learning rate will decrease at every call to
        `step`, regardless of the metric value

        `patience = 1`: learning rate is reduced as soon `step` is called
        with a metric value which does not improve over the best encountered

        `patience = 10`: learning rate is reduced if no improvement in the
        metric is recorded in 10 successive calls to `step`

    Parameters
    ----------
    objective
        String, can either be `"min"` or `"max"`
    patience
        The patience to observe before reducing the learning rate, nonnegative
        integer.
    base_lr
        Initial learning rate to be used.
    decay_factor
        Factor (between 0 and 1) by which to decrease the learning rate.
    min_learning_rate
        Lower bound for the learning rate, learning rate will never go below
        `min_learning_rate`.
    """

    patience: Patience
    learning_rate: float = field(default=0.01, metadata={"gt": 0})
    decay_factor: float = field(default=0.5, metadata={"gt": 0, "lt": 1})
    min_learning_rate: float = 0.0
    max_num_decays: Optional[int] = None

    def __post_init__(self) -> None:
        assert self.learning_rate > self.min_learning_rate

    def step(self, metric_value: float) -> bool:
        """
        Inform the scheduler of the new value of the metric that is being
        optimized. This method should be invoked at regular intervals (e.g. at
        the end of every epoch, after computing a validation score).

        Parameters
        ----------
        metric_value
            Value of the metric that is being optimized.

        Returns
        -------
        bool value indicating, whether to continue training
        """

        self.patience.step(metric_value)

        should_continue = True

        if self.patience.exceeded or not np.isfinite(metric_value):
            if (
                self.learning_rate == self.min_learning_rate
                or self.max_num_decays is not None
                and self.max_num_decays <= self.patience.num_resets
            ):
                should_continue = False

            # Even though we ask not to continue, we still reset the patience
            # because we might still end up continuing training. (Can Happen
            # in testing).
            self.patience.reset()

            self.learning_rate *= self.decay_factor
            # ensure that we don't go below the minimum learning rate
            if self.learning_rate < self.min_learning_rate:
                self.learning_rate = self.min_learning_rate

        return should_continue


class LearningRateReduction(Callback):
    """
    This Callback decreases the learning rate based on the value of some
    validation metric to be optimized (maximized or minimized). The value of
    such metric is provided by calling the `step` method on the scheduler. A
    `patience` parameter must be provided, and the scheduler will reduce the
    learning rate if no improvement in the metric is done before `patience`
    observations of the metric.

    Examples:

        `patience = 0`: learning rate will decrease at every call to
        `step`, regardless of the metric value

        `patience = 1`: learning rate is reduced as soon `step` is called
        with a metric value which does not improve over the best encountered

        `patience = 10`: learning rate is reduced if no improvement in the
        metric is recorded in 10 successive calls to `step`

    Parameters
    ----------
    objective
        String, can either be `"min"` or `"max"`.
    patience
        The patience to observe before reducing the learning rate, nonnegative
        integer.
    base_lr
        Initial learning rate to be used.
    decay_factor
        Factor (between 0 and 1) by which to decrease the learning rate.
    min_lr
        Lower bound for the learning rate, learning rate will never go below
        `min_lr`.
    """

    @validated()
    def __init__(
        self,
        objective: Literal["min", "max"],
        patience: int,
        base_lr: float = 0.01,
        decay_factor: float = 0.5,
        min_lr: float = 0.0,
    ) -> None:
        assert (
            0 < decay_factor < 1
        ), "The value of `decay_factor` should be in the (0, 1) range"
        assert patience >= 0, "The value of `patience` should be >= 0"
        assert (
            0 <= min_lr <= base_lr
        ), "The value of `min_lr` should be >= 0 and <= base_lr"

        self.lr_scheduler = MetricAttentiveScheduler(
            patience=Patience(
                patience=patience, objective=Objective.from_str(objective)
            ),
            learning_rate=base_lr,
            decay_factor=decay_factor,
            min_learning_rate=min_lr,
        )

    def on_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
        best_epoch_info: Dict[str, Any],
        ctx: mx.Context,
    ) -> bool:
        should_continue = self.lr_scheduler.step(metric_value=epoch_loss)
        trainer.optimizer.set_learning_rate(self.lr_scheduler.learning_rate)
        return should_continue



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Tuple

import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .distribution import Distribution, _sample_multiple, getF, softplus
from .distribution_output import DistributionOutput


class LogitNormal(Distribution):
    r"""
    The logit-normal distribution.

    Parameters
    ----------
    mu
        Tensor containing the location, of shape
        `(*batch_shape, *event_shape)`.
    sigma
        Tensor indicating the scale, of shape `(*batch_shape, *event_shape)`.
    F
    """

    @validated()
    def __init__(self, mu: Tensor, sigma: Tensor) -> None:
        super().__init__()
        self.mu = mu
        self.sigma = sigma

    @property
    def F(self):
        return getF(self.mu)

    @property
    def batch_shape(self) -> Tuple:
        return self.mu.shape

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def event_dim(self) -> int:
        return 0

    def log_prob(self, x: Tensor) -> Tensor:
        F = self.F
        x_clip = 1e-3
        x = F.clip(x, x_clip, 1 - x_clip)
        log_prob = -1.0 * (
            F.log(self.sigma)
            + F.log(F.sqrt(2 * F.full(1, np.pi)))
            + F.log(x)
            + F.log(1 - x)
            + (
                (F.log(x) - F.log(1 - x) - self.mu) ** 2
                / (2 * (self.sigma**2))
            )
        )
        return log_prob

    def sample(self, num_samples=None, dtype=np.float32):
        def s(mu):
            F = self.F
            q_min = 1e-3
            q_max = 1 - q_min
            sample = F.sample_uniform(
                F.ones_like(mu) * F.full(1, q_min),
                F.ones_like(mu) * F.full(1, q_max),
            )
            transf_sample = self.quantile(sample)
            return transf_sample

        mult_samp = _sample_multiple(s, self.mu, num_samples=num_samples)
        return mult_samp

    def quantile(self, level: Tensor) -> Tensor:
        F = self.F
        exp = F.exp(
            self.mu
            + (self.sigma * F.sqrt(F.full(1, 2)) * F.erfinv(2 * level - 1))
        )
        return exp / (1 + exp)

    @property
    def args(self) -> List:
        return [self.mu, self.sigma]


class LogitNormalOutput(DistributionOutput):
    args_dim: Dict[str, int] = {"mu": 1, "sigma": 1}
    distr_cls: type = LogitNormal

    @classmethod
    def domain_map(cls, F, mu, sigma):
        sigma = F.maximum(softplus(F, sigma), cls.eps())
        return mu.squeeze(axis=-1), sigma.squeeze(axis=-1)

    @property
    def event_shape(self) -> Tuple:
        return ()

    def distribution(
        self, distr_args, loc=None, scale=None, **kwargs
    ) -> Distribution:
        return self.distr_cls(*distr_args)
